//
//  TESingleDocument.swift
//  TextEditDoc
//
//  Created by narita on 2017/12/25.
//  Copyright © 2017年 narita. All rights reserved.
//

import Cocoa

class TESingleDocument: NSDocument
{

    // MARK: - class methods

    override class var autosavesInPlace: Bool
    {        
        return true
    }

    // MARK: - init methods

    override init()
    {
        super.init()
        // Add your subclass-specific initialization here.
        
        self.textStorage =  NSTextStorage()
    }

    public convenience init(contentsOf url: URL, ofType typeName: String) throws
    {
        self.init()
        
        try self.read(from: url, ofType: typeName)
        
        // typeNameがOPMLとそれ以外で分岐
        if NSWorkspace.shared.type("org.opml.opml", conformsToType: typeName)
        {
            self.fileURL = nil
            self.fileType = kUTTypeRTF as String
            
            self.fileModificationDate = Date()
            
            // ファイルタイプを変更したので、"Undo不可能な編集"をした事にする
            self.updateChangeCount(NSDocument.ChangeType.changeDone)
            
            self.isDraft = true
        }
        else
        {
            self.fileURL = url
            self.fileType = typeName
            
            let theAttr = try FileManager.default.attributesOfFileSystem(forPath: url.path)
            
            self.fileModificationDate = theAttr[FileAttributeKey.modificationDate] as? Date
        }
    }

    // MARK: - dealloc
    
    // MARK: - NSCopying, hash, isEqual:
    
    // MARK: - NSCoding methods
    
    override func fileWrapper(ofType typeName: String) throws -> FileWrapper
    {
        // self.textStorageがnilならばエラー
        if self.textStorage == nil
        {
            NSException(
                name: NSExceptionName.objectNotAvailableException,
                reason: "self.textStorage is nil.").raise()
        }

        
        let theWorkspace = NSWorkspace.shared
        let theRange: NSRange  = NSMakeRange(0, self.textStorage!.length)
        
        if theWorkspace.type(kUTTypePlainText as String, conformsToType: typeName)
        {
            self.isRichText = false
            
            return
                try self.textStorage!.fileWrapper(from: theRange,
                                                  documentAttributes:
                    [NSAttributedString.DocumentAttributeKey.documentType : NSAttributedString.DocumentType.plain])

        }
        else if theWorkspace.type(kUTTypeRTF as String, conformsToType: typeName)
        {
            self.isRichText = true

            return
                try self.textStorage!.fileWrapper(from: theRange,
                                                  documentAttributes:
                    [NSAttributedString.DocumentAttributeKey.documentType : NSAttributedString.DocumentType.rtf])
        }
        else if theWorkspace.type(kUTTypeRTFD as String, conformsToType: typeName)
        {
            self.isRichText = true

            return
                try self.textStorage!.fileWrapper(from: theRange,
                                                  documentAttributes:
                    [NSAttributedString.DocumentAttributeKey.documentType : NSAttributedString.DocumentType.rtfd])
        }
        else if theWorkspace.type(kUTTypeHTML as String, conformsToType: typeName)
        {
            self.isRichText = true
            
            // fileWrapper(from:, documentAttributes:)を使うとなぜがパッケージ形式になってしまう。
            // NSDataを作ってから、FileWrapperにする
            let theData = try self.textStorage!.data(from: theRange,
                                                     documentAttributes:
                [NSAttributedString.DocumentAttributeKey.documentType : NSAttributedString.DocumentType.html])

            return FileWrapper(regularFileWithContents: theData)
        }
        else if theWorkspace.type(kUTTypeWebArchive as String, conformsToType: typeName)
        {
            self.isRichText = true
            
            return
                try self.textStorage!.fileWrapper(from: theRange,
                                                  documentAttributes:
                    [NSAttributedString.DocumentAttributeKey.documentType : NSAttributedString.DocumentType.webArchive])
        }
        else
        {
            NSException(
                name: NSExceptionName.invalidArgumentException,
                reason: "\(typeName) is not a recognized document type.").raise()
        }
        
        return FileWrapper()
    }
    
    override func read(from fileWrapper: FileWrapper, ofType typeName: String) throws
    {
        let theWorkspace = NSWorkspace.shared
        
        if theWorkspace.type(kUTTypePlainText as String, conformsToType: typeName)
        {
            self.textStorage = try NSTextStorage(data: fileWrapper.regularFileContents!,
                                                 options: [:],
                                                 documentAttributes: nil)
            self.isRichText = false
        }
        else if theWorkspace.type(kUTTypeRTF as String, conformsToType: typeName)
        {
            self.textStorage = try NSTextStorage(data: fileWrapper.regularFileContents!,
                                                 options: [:],
                                                 documentAttributes: nil)
            self.isRichText = true
        }
        else if theWorkspace.type(kUTTypeRTFD as String, conformsToType: typeName)
        {
            // このファイル形式は他のとtextStorageの生成方式が違う
            self.textStorage = NSTextStorage(rtfdFileWrapper: fileWrapper,
                                             documentAttributes: nil)
            self.isRichText = true
        }
        else if theWorkspace.type(kUTTypeHTML as String, conformsToType: typeName)
        {
            self.textStorage = try NSTextStorage(data: fileWrapper.regularFileContents!,
                                                 options: [:],
                                                 documentAttributes: nil)
            self.isRichText = true
        }
        else if theWorkspace.type(kUTTypeWebArchive as String, conformsToType: typeName)
        {
            self.textStorage = try NSTextStorage(data: fileWrapper.regularFileContents!,
                                                 options: [:],
                                                 documentAttributes: nil)
            self.isRichText = true
        }
        else if theWorkspace.type( "org.opml.opml", conformsToType: typeName)
        {
            // OPMLからTextを構築してゆく
            let theXMLDoc: XMLDocument = try XMLDocument(data: fileWrapper.regularFileContents!)
            let theXMLNodes = try theXMLDoc.nodes(forXPath: "opml/body/outline")
            
            var theText: String = ""
            
            func readFromXML(inXMLNodes: [XMLNode])
            {
                for i: XMLNode in inXMLNodes
                {
                    if let localText: String? = (i as! XMLElement).attribute(forName: "text")?.objectValue as! String!
                    {
                        theText.append( localText! )
                        theText.append( "\r" )
                    }
                    
                    // ここで、iにsubnodeがあれば再帰する
                    if i.children != nil
                    {
                        readFromXML(inXMLNodes: i.children! )
                    }
                }
            }
            
            readFromXML(inXMLNodes: theXMLNodes)
            
            self.textStorage = NSTextStorage(attributedString: NSAttributedString(string: theText))
            
            self.isRichText = false
        }
        else
        {
            NSException(
                name: NSExceptionName.invalidArgumentException,
                reason: "\(typeName) is not a recognized document type.").raise()
        }
    }

    // これは、TextEditの[Document writableTypesForSaveOperation]をswift風に書いただけ
    @objc override func writableTypes(for saveOperation: NSDocument.SaveOperationType) -> [String]
    {
        var theResult = TESingleDocument.writableTypes
                
        if saveOperation == .saveAsOperation
        {
            // リッチテキストの時は、PlaneTextでは保存できないようにする
            if self.isRichText
            {
                // テキストは削除する。
                theResult.remove(at: theResult.index(of: kUTTypePlainText as String)! )
            }
            
            // 画像などの添付があれば、rtfdかwebArchiveでしか保存できない
            if self.textStorage!.containsAttachments(in: NSMakeRange(0, self.textStorage!.length))
            {
                theResult = [kUTTypeRTFD as String, kUTTypeWebArchive as String]
            }
        }
        
        return theResult
    }
    
    
    // MARK: - restorableState methods
    
    // MARK: - life cycle methods
    
    // MARK: - action methods
    
    @IBAction func changeToRichText(_ sender: Any?)
    {
        self.isRichText = true
    }

    @IBAction func changeToNonRichText(_ sender: Any?)
    {
        let theString = self.textStorage?.string
        let theRange: NSRange  = NSMakeRange(0, self.textStorage!.length)

        self.textStorage?.replaceCharacters(in: theRange, with: theString!)

        self.isRichText = false
    }
    
    // MARK: - event handling methods
    
    // MARK: - drawing methods
    
    // MARK: - delegate/datasource methods
    
    // MARK: - accessor methods (in pairs)
    
    @objc var textStorage: NSTextStorage?
    @objc dynamic var isRichText: Bool = false

    // MARK: - Utility methods

    override func makeWindowControllers()
    {
        // Returns the Storyboard that contains your Document window.
        let storyboard = NSStoryboard(name: NSStoryboard.Name("TEWindowController"), bundle: nil)
        
        let windowController = storyboard.instantiateInitialController() as! TEWindowController

        self.addWindowController(windowController)
    }

}

